package jp.sourceforge.csvparser;

import java.io.Reader;

import junit.framework.TestCase;

public class BasicCSVParserTest extends TestCase {

    private BasicCSVParser parser;

    public BasicCSVParserTest(String name) {

        super(name);
    }

    protected void setUp() throws Exception {

        super.setUp();
        parser = new BasicCSVParser();
    }

    private CSVHandler createCSVHandler() {

        return new MockCSVHandler();
    }

    public void testEmpty() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCRChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\ra");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "endRecord",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCRDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\r,");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "endRecord",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCRQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\r\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "endRecord",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCRCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\r\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "endRecord",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCRCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\r\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "endRecord",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a,");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "character:\"",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharCRChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a\ra");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharCRDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a\r,");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharCRQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a\r\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharCRCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a\r\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharCRCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("a\r\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testCharChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("aa");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",,");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimCRChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",\ra");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimCRDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",\r,");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimCRQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",\r\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimCRCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",\r\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimCRCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",\r\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testDelimChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader(",a");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\",");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:,",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\r",
                "character:\n",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\n",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\r",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotCRChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\ra");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\r",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotCRDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\r,");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\r",
                "character:,",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotCRQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\r\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\r",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotCRCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\r\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\r",
                "character:\r",
                "character:\n",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotCRCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\r\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\r",
                "character:\r",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"a");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\",");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\"",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotCRChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"\ra");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotCRDelim() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"\r,");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotCRQuot() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"\r\"");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotCRCRLF() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"\r\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotCRCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"\r\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "endRecord",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotQuotChar() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\"a");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:a",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testTab() throws Exception {

        parser.setDelimiter('\t');
        assertEquals('\t', parser.getDelimiter());

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\t");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "endValue",
                "startValue",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testQuotTab() throws Exception {

        parser.setDelimiter('\t');
        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\"\t");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:\t",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testSimple() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("1,One,First\r\n2,Two,Second\r\n3,Three,Th\"i\"rd\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:1",
                "endValue",
                "startValue",
                "character:O",
                "character:n",
                "character:e",
                "endValue",
                "startValue",
                "character:F",
                "character:i",
                "character:r",
                "character:s",
                "character:t",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "character:2",
                "endValue",
                "startValue",
                "character:T",
                "character:w",
                "character:o",
                "endValue",
                "startValue",
                "character:S",
                "character:e",
                "character:c",
                "character:o",
                "character:n",
                "character:d",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "character:3",
                "endValue",
                "startValue",
                "character:T",
                "character:h",
                "character:r",
                "character:e",
                "character:e",
                "endValue",
                "startValue",
                "character:T",
                "character:h",
                "character:\"",
                "character:i",
                "character:\"",
                "character:r",
                "character:d",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testComplex() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("1,\"One,First\"\r\n2,\"Two,Second\r\n3,Three,Th\"\"i\"\"rd\"\r\n");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:1",
                "endValue",
                "startValue",
                "character:O",
                "character:n",
                "character:e",
                "character:,",
                "character:F",
                "character:i",
                "character:r",
                "character:s",
                "character:t",
                "endValue",
                "endRecord",
                "startRecord",
                "startValue",
                "character:2",
                "endValue",
                "startValue",
                "character:T",
                "character:w",
                "character:o",
                "character:,",
                "character:S",
                "character:e",
                "character:c",
                "character:o",
                "character:n",
                "character:d",
                "character:\r",
                "character:\n",
                "character:3",
                "character:,",
                "character:T",
                "character:h",
                "character:r",
                "character:e",
                "character:e",
                "character:,",
                "character:T",
                "character:h",
                "character:\"",
                "character:i",
                "character:\"",
                "character:r",
                "character:d",
                "endValue",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }

    public void testEndRecord() throws Exception {

        MockCSVHandler handler = new MockCSVHandler() {
            
            public boolean endRecord() {
                
                super.endRecord();
                return true;
            }
        };

        int ch = 0;
        Reader reader = CSVTestUtils.openReader("1,One,First\r\n2,Two,Second\r\n3,Three,Third\r\n");
        
        try {
            parser.parse(reader, handler);
            ch = reader.read();
            
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "startValue",
                "character:1",
                "endValue",
                "startValue",
                "character:O",
                "character:n",
                "character:e",
                "endValue",
                "startValue",
                "character:F",
                "character:i",
                "character:r",
                "character:s",
                "character:t",
                "endValue",
                "endRecord",
        };
        CSVTestUtils.assertEquals(expected, actual);
        assertEquals('2', ch);
    }

    public void testLFCRLFCR() throws Exception {

        MockCSVHandler handler = new MockCSVHandler();
        Reader reader = CSVTestUtils.openReader("\n\r\n\r");
        
        try {
            parser.parse(reader, handler);
        } finally {
            reader.close();
        }

        String[] actual = handler.getCalls();
        String[] expected = {
                "startDocument",
                "startRecord",
                "endRecord",
                "startRecord",
                "endRecord",
                "startRecord",
                "endRecord",
                "endDocument"
        };
        CSVTestUtils.assertEquals(expected, actual);
    }
}